// Here Tree is now a class the protects its private representation and thus
// enforces its invariants.
#pragma once

#include <memory>
#include <stdexcept>
#include <string>
#include <vector>

namespace bintree {

class Tree {
public:
    Tree();

    bool empty() const;
    size_t size() const;
    bool member(const std::string&) const;
    unsigned lookup(const std::string&) const;
    std::vector<std::string> get_keys() const;

    void insert(const std::string&, unsigned);

private:
    struct Node;
    using link_t = std::shared_ptr<Node>;

    // Static functions are functions that don't operate on an instance of
    // the class of which they are members. That is, these three functions do
    // not get a Tree when they are called:
    static Tree::link_t find_subtree(Tree::link_t, const std::string&);
    static Tree::link_t& find_subtree_mut(Tree::link_t&, const std::string&);
    static void get_keys_helper(Tree::link_t, std::vector<std::string>&);

    link_t root_;
    size_t size_;
};

// An exception thrown by lookup(Tree, const std::string&) when the key isn't
// found.
struct Not_found : std::runtime_error
{
    Not_found(const std::string& key) :
            runtime_error("key not found: " + key) { }
};

}